@@ -37,8 +37,8 @@
  * ncurses and SLang seem to send different characters when the Enter key is
  * pressed, so define some macros to properly detect the Enter key.
  */
-#define M_ENTER_C '\r'
-#define M_ENTER_S "\r"
+#define MUTT_ENTER_C '\r'
+#define MUTT_ENTER_S "\r"
 
 #else /* USE_SLANG_CURSES */
 
@@ -52,8 +52,8 @@
 # include <curses.h>
 #endif
 
-#define M_ENTER_C '\n'
-#define M_ENTER_S "\n"
+#define MUTT_ENTER_C '\n'
+#define MUTT_ENTER_S "\n"
 
 #endif /* USE_SLANG_CURSES */
 
@@ -64,8 +64,8 @@
 #undef lines
 #endif /* lines */
 
-#define CLEARLINE(x) move(x,0), clrtoeol()
-#define CENTERLINE(x,y) move(y, (COLS-strlen(x))/2), addstr(x)
+#define CLEARLINE(win,x) mutt_window_clearline(win, x)
+#define CENTERLINE(win,x,y) mutt_window_move(win, y, (win->cols-strlen(x))/2), addstr(x)
 #define BEEP() do { if (option (OPTBEEP)) beep(); } while (0)
 
 #if ! (defined(USE_SLANG_CURSES) || defined(HAVE_CURS_SET))
@@ -77,7 +77,6 @@ void mutt_curs_set (int);
 #else
 #define mutt_curs_set(x)
 #endif
-#define PAGELEN (LINES-3)
 
 #define ctrl(c) ((c)-'@')
 
@@ -124,6 +123,14 @@ enum
   MT_COLOR_UNDERLINE,
   MT_COLOR_INDEX,
   MT_COLOR_PROMPT,
+#ifdef USE_SIDEBAR
+  MT_COLOR_DIVIDER,
+  MT_COLOR_FLAGGED,
+  MT_COLOR_HIGHLIGHT,
+  MT_COLOR_NEW,
+  MT_COLOR_SB_INDICATOR,
+  MT_COLOR_SB_SPOOLFILE,
+#endif
   MT_COLOR_MAX
 };
 
@@ -139,8 +146,8 @@ typedef struct color_line
   struct color_line *next;
 } COLOR_LINE;
 
-#define M_PROGRESS_SIZE		(1<<0)	/* traffic-based progress */
-#define M_PROGRESS_MSG		(1<<1)	/* message-based progress */
+#define MUTT_PROGRESS_SIZE      (1<<0)  /* traffic-based progress */
+#define MUTT_PROGRESS_MSG       (1<<1)  /* message-based progress */
 
 typedef struct
 {
@@ -161,14 +168,43 @@ void mutt_progress_init (progress_t* progress, const char *msg,
  * was initialized with positive size, otherwise no percentage is shown */
 void mutt_progress_update (progress_t* progress, long pos, int percent);
 
-static inline int mutt_term_width(short wrap)
+/* Windows for different parts of the screen */
+typedef struct
+{
+  int rows;
+  int cols;
+  int row_offset;
+  int col_offset;
+} mutt_window_t;
+
+extern mutt_window_t *MuttHelpWindow;
+extern mutt_window_t *MuttIndexWindow;
+extern mutt_window_t *MuttStatusWindow;
+extern mutt_window_t *MuttMessageWindow;
+#ifdef USE_SIDEBAR
+extern mutt_window_t *MuttSidebarWindow;
+#endif
+
+void mutt_init_windows (void);
+void mutt_free_windows (void);
+void mutt_reflow_windows (void);
+int mutt_window_move (mutt_window_t *, int row, int col);
+int mutt_window_mvaddch (mutt_window_t *, int row, int col, const chtype ch);
+int mutt_window_mvaddstr (mutt_window_t *, int row, int col, const char *str);
+int mutt_window_mvprintw (mutt_window_t *, int row, int col, const char *fmt, ...);
+void mutt_window_clrtoeol (mutt_window_t *);
+void mutt_window_clearline (mutt_window_t *, int row);
+void mutt_window_getyx (mutt_window_t *, int *y, int *x);
+
+
+static inline int mutt_window_wrap_cols(mutt_window_t *win, short wrap)
 {
   if (wrap < 0)
-    return COLS > -wrap ? COLS + wrap : COLS;
+    return win->cols > -wrap ? win->cols + wrap : win->cols;
   else if (wrap)
-    return wrap < COLS ? wrap : COLS;
+    return wrap < win->cols ? wrap : win->cols;
   else
-    return COLS;
+    return win->cols;
 }
 
 extern int *ColorQuote;
