@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000,2003 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 1996-2000,2003,2013 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -66,7 +66,7 @@ static ADDRESS *result_to_addr (QUERY *r)
   if(!tmp->next && !tmp->personal)
     tmp->personal = safe_strdup (r->name);
   
-  mutt_addrlist_to_idna (tmp, NULL);
+  mutt_addrlist_to_intl (tmp, NULL);
   return tmp;
 }
 
@@ -183,7 +183,7 @@ static int query_search (MUTTMENU *m, regex_t *re, int n)
   return REG_NOMATCH;
 }
 
-static const char * query_format_str (char *dest, size_t destlen, size_t col,
+static const char * query_format_str (char *dest, size_t destlen, size_t col, int cols,
 				      char op, const char *src,
 				      const char *fmt, const char *ifstring,
 				      const char *elsestring,
@@ -193,7 +193,7 @@ static const char * query_format_str (char *dest, size_t destlen, size_t col,
   QUERY *query = entry->data;
   char tmp[SHORT_STRING];
   char buf2[STRING] = "";
-  int optional = (flags & M_FORMAT_OPTIONAL);
+  int optional = (flags & MUTT_FORMAT_OPTIONAL);
 
   switch (op)
   {
@@ -230,9 +230,9 @@ static const char * query_format_str (char *dest, size_t destlen, size_t col,
   }
 
   if (optional)
-    mutt_FormatString (dest, destlen, col, ifstring, query_format_str, data, 0);
-  else if (flags & M_FORMAT_OPTIONAL)
-    mutt_FormatString (dest, destlen, col, elsestring, query_format_str, data, 0);
+    mutt_FormatString (dest, destlen, col, cols, ifstring, query_format_str, data, 0);
+  else if (flags & MUTT_FORMAT_OPTIONAL)
+    mutt_FormatString (dest, destlen, col, cols, elsestring, query_format_str, data, 0);
 
   return src;
 }
@@ -242,8 +242,8 @@ static void query_entry (char *s, size_t slen, MUTTMENU *m, int num)
   ENTRY *entry = &((ENTRY *) m->data)[num];
 
   entry->data->num = num;
-  mutt_FormatString (s, slen, 0, NONULL (QueryFormat), query_format_str,
-		     (unsigned long) entry, M_FORMAT_ARROWCURSOR);
+  mutt_FormatString (s, slen, 0, MuttIndexWindow->cols, NONULL (QueryFormat), query_format_str,
+		     (unsigned long) entry, MUTT_FORMAT_ARROWCURSOR);
 }
 
 static int query_tag (MUTTMENU *menu, int n, int m)
